/****************************************************************************
** $Id: glbox.cpp,v 1.5.2.1 1999/09/08 08:30:38 aavit Exp $
**
** Implementation of GLWindow
** This is a simple QGLWidget displaying an openGL wireframe box
**
** The OpenGL code is mostly borrowed from Brian Pauls "spin" example
** in the Mesa distribution
**
****************************************************************************/

#include  <stdlib.h>
#include  <stdio.h>
#include  <iostream>
#include  <ctype.h>
#include  <math.h>
#include  <time.h>
#include  <qapplication.h>
#include  <assert.h>
#include  <GL/glut.h>
#include  <fstream>
//#include  "ClusterTool.hpp"
#include  <GL/gl.h>

#include  "_generic.h"
//#include  "geometry.h"
#include  "real.h"
#include  "glwindow.h"
#include  "glwindow_geometry.h"

// Camera - define the camera view
class Camera
{
public:

    double fovy, aspect;
    double znear, zfar;
    gpoint3_t from, at, up;

    void look_at( const GBBox & bb, double ratio = 1.333333 );

    void apply();
    void apply(GLenum mode);
};


// ArcBall -  based on public domain code published by
//        Ken Shoemake in Graphics Gems IV.
struct BallData;

class Arcball 
{
private:
    int mousex, mousey;
    int startx, starty;
    BallData *ball;
    GPoint4  center;
    float radius;

    enum {NoDrag=0, Spin, Dolly, Zoom} dragging;

public:
    GPoint4  vNow;
    float cam[3];

public:
    Arcball() { ball=NULL; }
    Arcball(const GBBox  & bb, float aspect)
      { ball=NULL; init( bb,aspect); }
    ~Arcball();
    void init(const GBBox& bounds, float aspect);


    bool mouse_down(int which, int x, int y,
                    int  width, int  height );
    bool mouse_up(int which, int x, int y);
    bool motion(int x, int y, int  width, int  height );
    //    bool motion(int x, int y);

    void apply();
    void unapply();

    double *get_qnow();
};


/*!
  Create a GLWindow widget
*/

GLWindow::GLWindow( QWidget* parent, const char* name )
    : QGLWidget( parent, name )
{
    gl_scene = 0; //object = 0;
    f_draw_direct = false; //true;

    f_display_scene = true;

    camera = new Camera();
    arcball = new Arcball();
}


/*!
  Release allocated resources
*/

GLWindow::~GLWindow()
{
    glDeleteLists( gl_scene, 1 );
}


void  GLWindow::prepareGL()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    /*
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    camera->apply();
    */
    
    //glMatrixMode(GL_MODELVIEW);
    //arcball->apply();
}


void   GLWindow::smartDrawScene()
{
    if  ( f_draw_direct ) {
        drawScene();
        return;
    }
    if  ( gl_scene != 0 ) {
        glCallList( gl_scene );
        //printf( "calling list!\n" );
        return;
    }

    //printf( "1\n" );
    gl_scene = glGenLists(1);
    glNewList( gl_scene, GL_COMPILE);
    drawScene();
    glEndList();

    glCallList( gl_scene );
    //printf( "2\n" );
}



void GLWindow::paintGL()
{
    //glClear( GL_COLOR_BUFFER_BIT );
    prepareGL();
    glMatrixMode(GL_MODELVIEW);
    arcball->apply();

    //glCallList( object );

    //printf ( "%d, %d\n", f_display_terrain, f_display_simp );
    if  ( f_display_scene ) 
        smartDrawScene();

    glFlush();

    glMatrixMode(GL_MODELVIEW);
    arcball->unapply();
}


/*!
  Set up the OpenGL rendering state, and define display list
*/

void GLWindow::initializeGL()
{
    //printf( "initializeGL!\n" );

    //glCullFace( GL_FRONT_AND_BACK ); 
    glDisable( GL_CULL_FACE );
    //glEnable( GL_CULL_FACE );

    //glEnable( GL_LIGHTING );
    //glEnable( GL_LIGHT0 );
    glLightModeli( GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE );
    glEnable( GL_DEPTH_TEST );
    glShadeModel( GL_SMOOTH );
    glEnable( GL_NORMALIZE );

    //
    // Initialize the general GL setup
    //
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glEnable(GL_NORMALIZE);
    glShadeModel(GL_SMOOTH);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);

    GLfloat mat_ambient[] = { 0.2f, 0.2f, 0.2f, 1.0f };
    GLfloat mat_diffuse[] = { 0.5f, 0.5f, 0.5f, 1.0f };
    GLfloat mat_specular[] = { 0.1f, 0.1f, 0.1f, 1.0f };
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, mat_ambient);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, mat_diffuse);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular);
    glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, 40);


    GLfloat light_pos[] = {0.0f, 0.0f, 1.0f, 0.0f};
    glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
    glEnable(GL_LIGHT0);

    GLfloat light_pos_1[] = {5.0f, 0.0f, 1.0f, 0.0f};
    glLightfv( GL_LIGHT1, GL_POSITION, light_pos_1 );
    glEnable( GL_LIGHT1 );

    glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
    glEnable(GL_COLOR_MATERIAL);
    glColor3f(0.5f, 0.5f, 0.5f);
    glClearColor(0.3f, 0.3f, 1.0f, 0.0f);
}



/*!
  Set up the OpenGL view port, matrix mode, etc.
*/

void GLWindow::resizeGL( int w, int h )
{
    prepareView();
}


/*!
  Set the rotation angle of the object to \e degrees around the X axis.
*/

void  GLWindow::deleteGLObject()
{
    if  ( gl_scene != 0 )
        glDeleteLists( gl_scene, 1 );
    gl_scene = 0;
}


inline int  button_trans( int  button )
{
    switch  ( button ) {
    case  Qt::LeftButton  :
        return  1;
    case  Qt::MidButton :
        return  2;
    case  Qt::RightButton :
        return  3;
    };
    return  button;
}


void   GLWindow::mousePressEvent ( QMouseEvent * event )
{
    //int  mouse_x, mouse_y, bwidth;
    
    arcball->mouse_down( button_trans( event->button() ), 
                         event->x(),
                         event->y(),
                         width(), height() );
    //                               where[0], where[1]);
    repaint();
}


void   GLWindow::mouseReleaseEvent ( QMouseEvent * event )
{
    //int  mouse_x, mouse_y, bwidth;
    
    arcball->mouse_up( button_trans(event->button()), 
                       event->x(),
                       event->y() );
}


void   GLWindow::mouseMoveEvent ( QMouseEvent * event )
{
    bool  f_update;

    //int  mouse_x, mouse_y, bwidth;
    //printf( "mousemoveevent\n" );
    f_update = arcball->motion( event->x(),
                                event->y(),
                                width(), height() );
    if  ( f_update )
        repaint();
    //                               where[0], where[1]);
}


void   GLWindow::prepareView()
{
    //printf( "prepare view!\n");
    //deleteGLObject();

    GBBox  bb;
    getBBox( &bb );
    //bb.dump();

    glViewport( 0, 0, width(), height() );

    double aspect = (double)width()/(double)height();

    camera->look_at( bb, aspect );
    //camera->apply();

    arcball->init( bb, aspect );

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    camera->apply();
    //glMatrixMode(GL_MODELVIEW);

    //repaint();
}


void Camera::look_at(const GBBox & bb, double ratio)
{
    double d = 3* bb.get_radius() / tan(60*M_PI/180.0);

    bb.center( at );
    
    memcpy( from, at, sizeof( from ) );

    from[Z] += d;

    up[X] = 0; up[Y] = 1; up[Z] = 0;

    fovy = 60.0;
    aspect = ratio;

    znear = d/20;
    zfar = 10*d;
}

void Camera::apply()
{
    gluPerspective(fovy, aspect, znear, zfar);
    gluLookAt(from[X], from[Y], from[Z],
              at[X], at[Y], at[Z],
              up[X], up[Y], up[Z]);
}

void Camera::apply(GLenum mode)
{
    GLint last_mode; 
    
    glGetIntegerv(GL_MATRIX_MODE, &last_mode);
    glMatrixMode(mode);
    apply();
    glMatrixMode((GLenum)last_mode);
}


/************************************************************************
 "Stolen" from 

  Arcball rotation control -- The Arcball is a handy manipulator for
  spinning objects in 3D.  This particular controller also supports
  some limited camera and light movements as well.

  Copyright (C) 1998 Michael Garland.  See "COPYING.txt" for details.
  
  The core arcball code was originally published in:

	"Arcball Rotation Control"
	by Ken Shoemake <shoemake@graphics.cis.upenn.edu>
	in "Graphics Gems IV", Academic Press, 1994.

  and was placed in the public domain.

  Originally written for IRIS GL, the OpenGL code is based on the
  modifications of Victor Ng-Thow-Hing <victorng@dgp.toronto.edu>.
  Over time, I've made various modifications including:

	- Porting from ANSI C to C++
	- Nominal integration with the MixKit library
	- Removed drawing code

  $Id: Arcball.cxx,v 1.19 1999/12/08 17:26:10 garland Exp $

 ************************************************************************/

class Mat4
{
private:
    GPoint4 row[4];

protected:

    inline void copy(const Mat4& m);
    inline GPoint4  col(int i) const
        { return GPoint4(row[0].pnt[i],
                           row[1].pnt[i],
                           row[2].pnt[i],
                           row[3].pnt[i] ); }

public:
    // Standard matrices
    static Mat4 I;
    static Mat4 zero;
    static Mat4 unit;
    /*
    static Mat4  zero() {
        Mat4  m;

        m.row[ 0 ] = m.row[ 1 ] = m.row[ 2 ] = m.row[ 3 ] =
            GPoint4( 0, 0, 0, 0 );

        return  m;
    }
    static Mat4  I() {
        Mat4  m;

        m.row[ 0 ] = GPoint4( 1, 0, 0, 0 );
        m.row[ 1 ] = GPoint4( 0, 1, 0, 0 );
        m.row[ 2 ] = GPoint4( 0, 0, 1, 0 );
        m.row[ 3 ] = GPoint4( 0, 0, 0, 1 );

        return  m;
    }
    */

    static Mat4 trans(double,double,double);
    static Mat4 scale(double,double,double);
    static Mat4 xrot(double theta); //
    static Mat4 yrot(double theta); // Angles are in radians
    static Mat4 zrot(double theta); //
    //
    // rotation about an axis given by a unit vector; angle in radians.
    static Mat4 rot(double theta, double x, double y, double z); 

    //
    // Viewing matrices
    static Mat4 persp(double fovy, double aspect,
		      double zmin=0.0, double zmax=0.0);

    static Mat4 lookat(double e_x, double e_y, double e_z,
		       double at_x, double at_y, double at_z,
		       double up_x, double up_y, double up_z);

    static Mat4 viewport(double w, double h);


    // Standard constructors
    Mat4() { copy(zero); }
    Mat4(const gpoint4_t& r0,const gpoint4_t& r1,
         const gpoint4_t& r2,const gpoint4_t& r3)
    { row[0] = r0; 
    row[1] = *(gpoint4_t *)&r1; 
    row[2] = *(gpoint4_t *)&r2; 
    row[3] = *(gpoint4_t *)&r3;
    }
    Mat4(const GPoint4  & r0,const GPoint4   & r1,
         const GPoint4  & r2,const GPoint4   & r3 )
    { row[0] = r0.pnt; 
    row[1] = *(gpoint4_t *)&r1; 
    row[2] = *(gpoint4_t *)&r2; 
    row[3] = *(gpoint4_t *)&r3;
    }
    Mat4(const Mat4& m) { copy(m); }

    // Access methods
    // M(i, j) == row i;col j
    double& operator()(int i, int j)       { return row[i].pnt[j]; }
    double  operator()(int i, int j) const { return row[i].pnt[j]; }
    gpoint4_t&       operator[](int i)       { return row[i].pnt; }
    const gpoint4_t& operator[](int i) const { return row[i].pnt; }

    operator       double*()       { return row[0].pnt; }
    operator const double*() const { return row[0].pnt; }

    // Comparison methods
    inline int operator==(const Mat4&);

    // Assignment methods
    inline Mat4& operator=(const Mat4& m) { copy(m); return *this; }
    inline Mat4& operator+=(const Mat4& m);
    inline Mat4& operator-=(const Mat4& m);

    inline Mat4& operator*=(double s);
    inline Mat4& operator/=(double s);


    // Arithmetic methods
    inline Mat4 operator+(const Mat4& m) const;
    inline Mat4 operator-(const Mat4& m) const;
    inline Mat4 operator-() const;

    inline Mat4 operator*(double s) const;
    inline Mat4 operator/(double s) const;
    Mat4 operator*(const Mat4& m) const;

    inline GPoint4 operator*(const GPoint4  & v) const; // [x y z w]
    inline GPoint3 operator*(const GPoint3  & v) const; // [x y z w]

    // Matrix operations
    double det() const;
    Mat4 transpose() const;
    Mat4 adjoint() const;
    double invert(Mat4&) const;
    double cramerInvert(Mat4&) const;
};

Mat4 Mat4::I(GPoint4(1,0,0,0),
             GPoint4(0,1,0,0),
             GPoint4(0,0,1,0),
             GPoint4(0,0,0,1));
Mat4 Mat4::zero(GPoint4(0,0,0,0),
                GPoint4(0,0,0,0),
                GPoint4(0,0,0,0),
                GPoint4(0,0,0,0));
Mat4 Mat4::unit(GPoint4(1,1,1,1),
                GPoint4(1,1,1,1),
                GPoint4(1,1,1,1),
                GPoint4(1,1,1,1));


inline void Mat4::copy(const Mat4& m)
{
    row[0] = m.row[0]; row[1] = m.row[1];
    row[2] = m.row[2]; row[3] = m.row[3];
}

inline int Mat4::operator==(const Mat4& m)
{
    return row[0].pnt==m.row[0].pnt &&
	   row[1].pnt==m.row[1].pnt &&
	   row[2].pnt==m.row[2].pnt &&
	   row[3].pnt==m.row[3].pnt ;
}

inline Mat4& Mat4::operator+=(const Mat4& m)
{
    row[0] += m.row[0]; row[1] += m.row[1];
    row[2] += m.row[2]; row[3] += m.row[3];
    return *this;
}

inline Mat4& Mat4::operator-=(const Mat4& m)
{
    row[0] -= m.row[0]; row[1] -= m.row[1];
    row[2] -= m.row[2]; row[3] -= m.row[3];
    return *this;
}

inline Mat4& Mat4::operator*=(double s)
{
    row[0] *= s; row[1] *= s; row[2] *= s; row[3] *= s;
    return *this;
}

inline Mat4& Mat4::operator/=(double s)
{
    row[0] /= s; row[1] /= s; row[2] /= s; row[3] /= s;
    return *this;
}

inline Mat4 Mat4::operator+(const Mat4& m) const
{
    return Mat4(row[0] + m.row[0],
		row[1] + m.row[1],
		row[2] + m.row[2],
		row[3] + m.row[3]);
}

inline Mat4 Mat4::operator-(const Mat4& m) const
{
    return Mat4(row[0]-m.row[0],
		row[1]-m.row[1],
		row[2]-m.row[2],
		row[3]-m.row[3]);
}

inline Mat4 Mat4::operator-() const
{
    return Mat4(-row[0], -row[1], -row[2], -row[3]);
}

inline Mat4 Mat4::operator*(double s) const
{
    return Mat4(row[0]*s, row[1]*s, row[2]*s, row[3]*s);
}

inline Mat4 Mat4::operator/(double s) const
{
    return Mat4(row[0]/s, row[1]/s, row[2]/s, row[3]/s);
}

inline GPoint4 Mat4::operator*(const GPoint4& v) const
{
    return GPoint4(row[0]*v, row[1]*v, row[2]*v, row[3]*v);
}
//
// Transform a homogeneous 3-vector and reproject into normal 3-space
//
inline GPoint3 Mat4::operator*(const GPoint3  & v) const
{
    GPoint4 u=GPoint4(v,1);
    double w=row[3].pnt*u;

    if(w==0.0)
	return GPoint3(row[0].pnt*u, row[1].pnt*u, row[2].pnt*u);
    else
	return GPoint3(row[0].pnt*u/w, row[1].pnt*u/w, row[2].pnt*u/w);
}

extern bool jacobi(const Mat4& m, GPoint4& vals, GPoint4 vecs[4]);

#ifdef MXGL_INCLUDED
inline void glGetMatrix(Mat4& m, GLenum which=GL_MODELVIEW_MATRIX)
{ Mat4 tmp;  glGetDoublev(which, &tmp(0,0));  m=tmp.transpose(); }

inline void glLoadMatrix(const Mat4& m)
{ Mat4 tmp = m.transpose();  glLoadMatrixd(&tmp(0,0)); }

inline void glMultMatrix(const Mat4& m)
{ Mat4 tmp = m.transpose();  glMultMatrixd(&tmp(0,0)); }
#endif



//#include "stdmix.h"
//#include "GL.h"
//#include "Arcball.h"

enum AxisSet {NoAxes, CameraAxes, BodyAxes, OtherAxes, NSets};
enum BallType { ArcBall, TrackBall };
typedef Mat4 HMatrix;



/* ***** BallAux.c ***** */

typedef  GPoint4  Quat;

Quat qOne(0,0,0,1);

/* Return quaternion product qL * qR.  Note: order is important!
 * To combine rotations, use the product Mul(qSecond, qFirst),
 * which gives the effect of rotating by qFirst then qSecond. */
Quat Qt_Mul(Quat& qL, Quat& qR)
{
    Quat qq;
    qq.pnt[W] = qL[W]*qR[W] - qL[X]*qR[X] - qL[Y]*qR[Y] - qL[Z]*qR[Z];
    qq.pnt[X] = qL[W]*qR[X] + qL[X]*qR[W] + qL[Y]*qR[Z] - qL[Z]*qR[Y];
    qq.pnt[Y] = qL[W]*qR[Y] + qL[Y]*qR[W] + qL[Z]*qR[X] - qL[X]*qR[Z];
    qq.pnt[Z] = qL[W]*qR[Z] + qL[Z]*qR[W] + qL[X]*qR[Y] - qL[Y]*qR[X];
    return (qq);
}


/* Construct rotation matrix from (possibly non-unit) quaternion.
 * Assumes matrix is used to multiply column vector on the left:
 * vnew = mat vold.  Works correctly for right-handed coordinate system
 * and right-handed rotations. */
HMatrix *Qt_ToMatrix(const Quat& q, HMatrix& out)
{
    double Nq = norm2(q);
    double s = (Nq > 0.0) ? (2.0 / Nq) : 0.0;

    double xs = q[X]*s,  ys = q[Y]*s,  zs = q[Z]*s;
    double wx = q[W]*xs, wy = q[W]*ys, wz = q[W]*zs;
    double xx = q[X]*xs, xy = q[X]*ys, xz = q[X]*zs;
    double yy = q[Y]*ys, yz = q[Y]*zs, zz = q[Z]*zs;

    out(X,X) = 1.0 - (yy + zz); out(Y,X) = xy + wz; out(Z,X) = xz - wy;
    out(X,Y) = xy - wz; out(Y,Y) = 1.0 - (xx + zz); out(Z,Y) = yz + wx;
    out(X,Z) = xz + wy; out(Y,Z) = yz - wx; out(Z,Z) = 1.0 - (xx + yy);
    out(X,W) = out(Y,W) = out(Z,W) = out(W,X) = out(W,Y) = out(W,Z) = 0.0;
    out(W,W) = 1.0;
    return ((HMatrix *)&out);
}

/* Return conjugate of quaternion. */
Quat Qt_Conj(Quat& q)
{
    Quat qq;
    qq.pnt[X] = -q[X];
    qq.pnt[Y] = -q[Y];
    qq.pnt[Z] = -q[Z];
    qq.pnt[W] = q[W];
    return (qq);
}

/* Return vector formed from components */
inline GPoint4 V3_(float x, float y, float z)
{
    return GPoint4(x, y, z, 0);
}

/* Return norm of v, defined as sum of squares of components */
inline double V3_Norm(const GPoint4& v)
{
    return ( v[X]*v[X] + v[Y]*v[Y] + v[Z]*v[Z] );
}

/* Return unit magnitude vector in direction of v */
GPoint4 V3_Unit(const GPoint4& v)
{
    static GPoint4 u(0, 0, 0, 0);
    float vlen = sqrt(V3_Norm(v));
    if (vlen != 0.0) {
	u.pnt[X] = v[X]/vlen;
        u.pnt[Y] = v[Y]/vlen; 
        u.pnt[Z] = v[Z]/vlen;
    }
    return (u);
}

/* Return version of v scaled by s */
inline GPoint4 V3_Scale(const GPoint4& v, float s)
{
    return GPoint4(s*v[X], s*v[Y], s*v[Z], v[W]);
}

/* Return negative of v */
inline GPoint4 V3_Negate(const GPoint4& v)
{
    return GPoint4(-v[X], -v[Y], -v[Z], 0);
}

/* Return sum of v1 and v2 */
inline GPoint4 V3_Add(const GPoint4& v1, const GPoint4& v2)
{
    return GPoint4(v1[X]+v2[X], v1[Y]+v2[Y], v1[Z]+v2[Z], 0);
}

/* Return difference of v1 minus v2 */
inline GPoint4 V3_Sub(const GPoint4& v1, const GPoint4& v2)
{
    return GPoint4(v1[X]-v2[X], v1[Y]-v2[Y], v1[Z]-v2[Z], 0);
}

/* Halve arc between unit vectors v0 and v1. */
GPoint4 V3_Bisect(const GPoint4& v0, const GPoint4& v1)
{
    GPoint4 v(0, 0, 0, 0);
    v = V3_Add(v0, v1);

    float Nv = V3_Norm(v);
    if (Nv < 1.0e-5) {
	v = V3_(0, 0, 1);
    } else {
	v = V3_Scale(v, 1/sqrt(Nv));
    }
    return (v);
}

/* Return dot product of v1 and v2 */
double V3_Dot(const GPoint4& v1, const GPoint4& v2)
{
    return (v1[X]*v2[X] + v1[Y]*v2[Y] + v1[Z]*v2[Z]);
}


/* Return cross product, v1 x v2 */
inline GPoint4 V3_Cross(const GPoint4& v1, const GPoint4& v2)
{
    return GPoint4(v1[Y]*v2[Z]-v1[Z]*v2[Y],
		v1[Z]*v2[X]-v1[X]*v2[Z],
		v1[X]*v2[Y]-v1[Y]*v2[X],
		0);
}



/* ***** Ball.h ***** */

typedef GPoint4 *ConstraintSet;
struct BallData {
    GPoint4 center;
    double radius, aspect;
    Quat qNow, qDown, qDrag;
    GPoint4 vNow, vDown, vFrom, vTo, vrFrom, vrTo;
    HMatrix mNow, mDown;
    int dragging;
    ConstraintSet sets[NSets];
    int setSizes[NSets];
    AxisSet axisSet;
    int axisIndex;
    BallType ballType;
};

/* Public routines */
extern BallData *Ball_Alloc();
extern void Ball_Free(BallData *);
extern void Ball_Init(BallData *ball, double aspect);
extern void Ball_Place(BallData *ball, GPoint4 center, double radius
		       /*, Quat orient*/);
extern void Ball_Mouse(BallData *ball, GPoint4 vNow);
extern void Ball_UseSet(BallData *ball, AxisSet axisSet);
extern void Ball_Update(BallData *ball, int al);
extern void Ball_Value(BallData *ball, HMatrix& mNow);
extern void Ball_BeginDrag(BallData *ball);
extern void Ball_EndDrag(BallData *ball);

extern int Ball_MouseOutsideSphere(BallData *ball, GPoint4 vMouse);
extern void Ball_SetBallType(BallData *ball, BallType bt);
/* Private routines */



/* ***** BallMath.c ***** */

/* Convert window coordinates to sphere coordinates. */
GPoint4 MouseOnSphere(GPoint4 mouse, GPoint4 ballCenter, double ballRadius)
{
    GPoint4 ballMouse;
    register double mag;
    ballMouse.pnt[X] = (mouse[X] - ballCenter[X]) / ballRadius;
    ballMouse.pnt[Y] = (mouse[Y] - ballCenter[Y]) / ballRadius;
    mag = ballMouse[X]*ballMouse[X] + ballMouse[Y]*ballMouse[Y];
    if (mag > 1.0) {
	register double scale = 1.0/sqrt(mag);
	ballMouse.pnt[X] *= scale; ballMouse.pnt[Y] *= scale;
	ballMouse.pnt[Z] = 0.0;
    } else {
	ballMouse.pnt[Z] = sqrt(1 - mag);
    }
    ballMouse.pnt[W] = 0.0;
    return (ballMouse);
}

/* Construct a unit quaternion from two points on unit sphere */
Quat Qt_FromBallPoints(GPoint4 from, GPoint4 to, BallType bt)
{
    Quat qu;
    qu.pnt[X] = from[Y]*to[Z] - from[Z]*to[Y];
    qu.pnt[Y] = from[Z]*to[X] - from[X]*to[Z];
    qu.pnt[Z] = from[X]*to[Y] - from[Y]*to[X];

    if (bt == ArcBall){
	qu.pnt[W] = from[X]*to[X] + from[Y]*to[Y] + from[Z]*to[Z];
    }
    else {
	float  a, s, mag;

	a = .5*acos(from[X]*to[X] + from[Y]*to[Y] + from[Z]*to[Z]);
	mag = sqrt(qu[X]*qu[X] + qu[Y]*qu[Y] + qu[Z]*qu[Z]);
	qu.pnt[W] = cos(a); s = sin(a);
	if (s != 0)
	  s /= mag;
	qu.pnt[X] *= s;  qu.pnt[Y] *= s; qu.pnt[Z] *= s;
    }

    return (qu);
}

/* Convert a unit quaternion to two points on unit sphere */
void Qt_ToBallPoints(Quat q, GPoint4 *arcFrom, GPoint4 *arcTo)
{
    double s;
    s = sqrt(q[X]*q[X] + q[Y]*q[Y]);
    if (s == 0.0) {
	*arcFrom = V3_(0.0, 1.0, 0.0);
    } else {
	*arcFrom = V3_(-q[Y]/s, q[X]/s, 0.0);
    }
    (*arcTo).pnt[X] = q[W]*(*arcFrom)[X] - q[Z]*(*arcFrom)[Y];
    (*arcTo).pnt[Y] = q[W]*(*arcFrom)[Y] + q[Z]*(*arcFrom)[X];
    (*arcTo).pnt[Z] = q[X]*(*arcFrom)[Y] - q[Y]*(*arcFrom)[X];
    if (q[W] < 0.0) *arcFrom = V3_(-(*arcFrom)[X], -(*arcFrom)[Y], 0.0);
}

/* Force sphere point to be perpendicular to axis. */
GPoint4 ConstrainToAxis(GPoint4& loose, GPoint4& axis)
{
    GPoint4 onPlane = V3_Sub(loose, V3_Scale(axis, V3_Dot(axis, loose)));
    register float norm = V3_Norm(onPlane);

    if (norm > 0.0) {
	if (onPlane[Z] < 0.0) onPlane = V3_Negate(onPlane);
	return ( V3_Scale(onPlane, 1/sqrt(norm)) );
    } /* else drop through */
    if (axis[Z] == 1) {
	onPlane = V3_(1.0, 0.0, 0.0);
    } else {
	onPlane = V3_Unit(V3_(-axis[Y], axis[X], 0.0));
    }
    return (onPlane);
}

/* Find the index of nearest arc of axis set. */
int NearestConstraintAxis(GPoint4& loose, GPoint4 *axes, int nAxes)
{
    GPoint4 onPlane;
    register float max, dot;
    register int i, nearest;
    max = -1; nearest = 0;
    for (i=0; i<nAxes; i++) {
	onPlane = ConstrainToAxis(loose, axes[i]);
	dot = V3_Dot(onPlane, loose);
	if (dot>max) {
	    max = dot; nearest = i;
	}
    }
    return (nearest);
}



/* ***** Ball.c ***** */
/* Ken Shoemake, 1993 */
/* Modified by Victor Ng, Jan. 1996 for OpenGL */

#define FALSE 0
#define TRUE 1

#define LG_NSEGS 4
#define NSEGS (1<<LG_NSEGS)
#define RIMCOLOR()    glColor3f(0.0f, 1.0f, 1.0f);
#define FARCOLOR()    glColor3f(0.8f, 0.5f, 0.0f);
#define NEARCOLOR()   glColor3f(1.0f, 0.8f, 0.0f);
#define DRAGCOLOR()   glColor3f(0.0f, 1.0f, 1.0f);
#define RESCOLOR()    glColor3f(0.8f, 0.0f, 0.0f);



GPoint4 otherAxis(-0.48, 0.80, 0.36, 1);

BallData *Ball_Alloc()
{
    return (BallData *)malloc( sizeof(BallData) );
}

void Ball_Free(BallData *ball)
{
    free(ball);
}

/* Establish reasonable initial values for controller. */
void Ball_Init(BallData *ball, double aspect)
{
    ball->center = qOne;
    ball->radius = 1.0;
    ball->aspect = aspect;
    ball->ballType = ArcBall;
    ball->vDown = ball->vNow = qOne;
    ball->qDown = ball->qNow = qOne;

    ball->mNow = Mat4::I;
    ball->mDown = Mat4::I;

    ball->dragging = FALSE;
    ball->axisSet = NoAxes;

    ball->sets[CameraAxes]     = (GPoint4 *)&Mat4::I(0,0);  // Mat4::I.row(0)
    ball->setSizes[CameraAxes] = 3;

    ball->sets[BodyAxes] = (GPoint4 *)&ball->mDown(0,0);  // mDown.row(0)
    ball->setSizes[BodyAxes] = 3;

    ball->sets[OtherAxes] = &otherAxis;
    ball->setSizes[OtherAxes] = 1;
}

/* Set the center and size of the controller. */
void Ball_Place(BallData *ball, GPoint4 center, double radius)
{
    ball->center = center;
    ball->radius = radius;
}

/* Incorporate new mouse position. */
void Ball_Mouse(BallData *ball, GPoint4 vNow)
{
    ball->vNow = vNow;
}

/* Choose a constraint set, or none. */
void Ball_UseSet(BallData *ball, AxisSet axisSet)
{
    /* if (!ball->dragging) */
	ball->axisSet = axisSet;
}

/* Using vDown, vNow, dragging, and axisSet, compute rotation etc. */
/* To use closest constraint axis, set aI to -1 */
void Ball_Update(BallData *ball, int aI)
{
    int setSize = ball->setSizes[ball->axisSet];
    GPoint4 *set = (GPoint4 *)(ball->sets[ball->axisSet]);
    ball->vFrom = MouseOnSphere(ball->vDown, ball->center, ball->radius);
    ball->vTo = MouseOnSphere(ball->vNow, ball->center, ball->radius);
    if (ball->dragging) {
        if (ball->axisSet!=NoAxes) {
	    aI = (aI==-1) ? ball->axisIndex : aI;
            ball->vFrom = ConstrainToAxis(ball->vFrom, set[aI]);
            ball->vTo = ConstrainToAxis(ball->vTo, set[aI]);
        }
        ball->qDrag =Qt_FromBallPoints(ball->vFrom, ball->vTo, ball->ballType);
        ball->qNow = Qt_Mul(ball->qDrag, ball->qDown);
    } else {
        if (ball->axisSet!=NoAxes) {
            ball->axisIndex = (aI==-1) ?
	                      NearestConstraintAxis(ball->vTo, set, setSize) :
	                      aI;
        }
    }
    Qt_ToBallPoints(ball->qDown, &ball->vrFrom, &ball->vrTo);
    Qt_ToMatrix(Qt_Conj(ball->qNow), ball->mNow); /* Gives transpose for GL. */
}

/* Returns TRUE iff vMouse is a point off the arcball sphere */
int Ball_MouseOutsideSphere(BallData *ball, GPoint4 vMouse)
{
    GPoint4  vTest = MouseOnSphere(vMouse, ball->center, ball->radius);

    return vTest[Z] == 0.0;
}

/* Set the BallType: 'ArcBall' rotates through _twice_ the angle
 * between the two points on the sphere, and exhibits no
 * hysteresis. 'TrackBall' rotates through exactly the angle between
 * the two points on the sphere, like a physical trackball, and
 * consequently exhibits hysteresis. */
void Ball_SetBallType(BallData *ball, BallType bt)
{
    ball->ballType = bt;
}

/* Return rotation matrix defined by controller use. */
void Ball_Value(BallData *ball, HMatrix& mNow)
{
    mNow = ball->mNow;
}


/* Begin drag sequence. */
void Ball_BeginDrag(BallData *ball)
{
    ball->dragging = TRUE;
    ball->vDown = ball->vNow;
}

/* Stop drag sequence. */
void Ball_EndDrag(BallData *ball)
{
    ball->dragging = FALSE;
    ball->qDown = ball->qNow;

    ball->mDown = ball->mNow;
}



Arcball::~Arcball()
{
    if( ball ) Ball_Free(ball);
}

void Arcball::init(const GBBox   & bounds, float aspect)
{
    if( ball ) Ball_Free(ball);  // Don't try to reuse old Balls
    ball = Ball_Alloc();
    Ball_Init(ball, aspect);

    Ball_UseSet(ball, NoAxes);
    Ball_SetBallType(ball, ArcBall);

    GPoint3  tmp_center;
    bounds.center( tmp_center );

    center = GPoint4( tmp_center, 1 );

    radius = bounds.get_radius();
    cam[X] = cam[Y] = cam[Z] = 0.0f;
    dragging = NoDrag;

    Ball_Place(ball, qOne, 0.9 /* 0.75 */);
}

bool Arcball::mouse_down(int button, int x, int y, 
                         int  width, int  height )
{
    mousex = startx = x;
    mousey = starty = y;

    //printf( "mouse_down( button: %d, x: %d, y: %d, width: %d, height: %d\n",
    //        button, x, y, width, height );
    if( button == 1 ) { 
        //int width, height;
        //glGetViewport(NULL, NULL, &width, &height);

        vNow.pnt[X] = 2.0*startx/(float)width-1.0;
        vNow.pnt[Y] = -2.0*starty/(float)height+1.0;

        Ball_Mouse(ball, vNow);

        Ball_BeginDrag(ball);

        dragging = Spin;
        //printf/( "spinning!\n" );
    }
    else if( button == 2 )
    {
        dragging = Dolly;
    }
    else if( button == 3 )
    {
        dragging = Zoom;
    }

    return true;
}

bool Arcball::mouse_up(int /*button*/, int /*x*/, int /*y*/)
{
    startx = mousex = 0;
    starty = mousey = 0;

    if( dragging == Spin )
        Ball_EndDrag(ball);

    dragging = NoDrag;

    return false;
}

bool Arcball::motion(int x, int y, int  width, int  height )
{
    if( !dragging )
        return false;

    if( dragging == Spin )
    {
        mousex = x;
        mousey = y;

        //int width, height;
        //glGetViewport(NULL, NULL, &width, &height);

        vNow.pnt[X] = 2.0*mousex/(float)width-1.0;
        vNow.pnt[Y] = -2.0*mousey/(float)height+1.0;

        Ball_Mouse(ball, vNow);
    }
    else if( dragging == Dolly )
    {
	//int width, height;
        //	glGetViewport(NULL, NULL, &width, &height);

	cam[X] += radius * (float)(x - mousex) / (float)width;
	cam[Y] += radius * (float)(mousey - y) / (float)height;

        mousex = x;
        mousey = y;
    }
    else if( dragging == Zoom )
    {
        cam[Z] += 0.02*radius*(float)(y - mousey);
        mousey = y;
    }

    return true;
}

void Arcball::apply()
{
    Ball_Update(ball, -1);

    glPushMatrix();
    glTranslatef(cam[X], cam[Y], cam[Z]);

    HMatrix mNow;
    Ball_Value(ball, mNow);

    glTranslatef(center[X], center[Y], center[Z]);
    glMultMatrixd((double *)&mNow);
    glTranslatef(-center[X], -center[Y], -center[Z]);
}

void Arcball::unapply()
{
    glPopMatrix();
}


/* glbox.C - End of File ------------------------------------------*/


