/*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
 * obb.C -
 *     
\*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*/

#include  <stdlib.h>
#include  <stdio.h>
#include  <assert.h>
#include  <math.h>
#include  <time.h>
#include  <string.h>

//#include  "dmalloc.h"

#include  "_generic.h"
#include  "sarray.h"
#include  "real.h"
#include  "point.h"
#include  "misc.h"

/*--- Constants ---*/


/*--- Start of Code ---*/

/* from MAGIC */
void EigenSymmetric3(double mat[3][3], double eval[3],
                     double evec[3][3]);

#define X 0
#define Y 1
#define Z 2

inline void  Zero(double pt[3]){
    pt[X] = 0.0;
    pt[Y] = 0.0;
    pt[Z] = 0.0;
}


static Point3d    compute_pnt( Point3d  * evecs,
                               double  * coff )
{
    Point3d  pnt;

    pnt.add_scale( evecs[ 0 ], coff[ 0 ] );
    pnt.add_scale( evecs[ 1 ], coff[ 1 ] );
    pnt.add_scale( evecs[ 2 ], coff[ 2 ] );

    return  pnt;
}


void  extractOrthonormalBase(Point3d  & p_a,
                             Point3d  & p_b,
                             Point3d  & p_c )
{
    Point3d  evec[ 3 ];

    evec[ 0 ] = p_a;
    evec[ 1 ] = p_b;
    evec[ 2 ] = p_c;

    if  ( evec[ 0 ].length() < evec[ 1 ].length() )
        exchange( evec[ 0 ], evec[ 1 ] );
    if  ( evec[ 0 ].length() < evec[ 2 ].length() )
        exchange( evec[ 0 ], evec[ 2 ] );
    if  ( evec[ 1 ].length() < evec[ 2 ].length() )
        exchange( evec[ 1 ], evec[ 2 ] );

    // The vectors are sorted by length...
    if  ( evec[ 0 ].length() == 0.0 ) {
        p_a = Point3d( 1, 0, 0 );
        p_b = Point3d( 0, 1, 0 );
        p_c = Point3d( 0, 0, 1 );
        return;
    }

    evec[ 0 ].normalize();
    //evec[ 1 ].add_scale( evec[ 0 ], - evec[ 0 ].dot_prod( evec[ 1 ] ) );

    evec[ 1 ].normalize();
    evec[ 2 ].normalize();
    
    //printf( "dot prod 0,1: %g\n", evec[ 0 ].dot_prod( evec[ 1 ] ) );

    //printf( "dot prod 0,2: %g\n", evec[ 0 ].dot_prod( evec[ 2 ] ) );
    evec[ 2 ].add_scale( evec[ 0 ], - evec[ 0 ].dot_prod( evec[ 2 ] ) );
    printf( "dot prod 0,2: %g\n", evec[ 0 ].dot_prod( evec[ 2 ] ) );

    evec[ 2 ].add_scale( evec[ 1 ], - evec[ 1 ].dot_prod( evec[ 2 ] ) );
    printf( "dot prod 1,2: %g\n", evec[ 1 ].dot_prod( evec[ 2 ] ) );

    evec[ 2 ].print();
    printf( "\n\n" );

    evec[ 2 ].normalize();

    // The vectors are sorted by length...
    if  ( evec[ 1 ].length() == 0.0 ) {
        assert( false );
        return;

    }
    if  ( ( evec[ 0 ].length() != 0.0 )
          &&  ( evec[ 1 ].length() != 0.0 ) 
          &&  ( evec[ 2 ].length() == 0.0 ) ) {
        evec[ 2 ] = cross_prod( evec[ 0 ], evec[ 1 ] );
        evec[ 2 ].normalize();
    }

    /*
    printf( "\n\n res:\n" );

    evec[ 0 ].print();
    printf( "\n" );
    evec[ 1 ].print();
    printf( "\n" );
    evec[ 2 ].print();
    printf( "\n" );
    */
    p_a = evec[ 0 ];
    p_b = evec[ 1 ];
    p_c = evec[ 2 ];
}




void  EigenSymetric3vec( double  mCoVars[3][3],
                         Point3d  & p_a,
                         Point3d  & p_b,
                         Point3d  & p_c )
{
    //double  XfTemp;
    Point3d  evec[ 3 ];
    double  EigenVals[3];
    double  mXForm[3][3];

    /* put eigenvectors into mXForm */
    //EigenSymmetric3(mCoVars, EigenVals, mXForm);
    Jacobi_EigenSymmetric3( mCoVars, EigenVals, mXForm );

    //printf( "eigen values: %g %g %g\n",
    //         EigenVals[ 0 ],
    //        EigenVals[ 1 ],
    //        EigenVals[ 2 ] );


    /* Normalize the eigenvectors to make unit axes. */
    for  ( int  i = 0; i < 3; i++ ) {
        //printf( "___%d: %g %g %g\n", i,
        //        mXForm[i][0],
        //        mXForm[i][1],
        //       mXForm[i][2] );

        evec[ i ] = Point3d( mXForm[i][0], 
                             mXForm[i][1],
                             mXForm[i][2] );
        evec[ i ].normalize_big();
    }

    p_a = evec[ 0 ];
    p_b = evec[ 1 ];
    p_c = evec[ 2 ];
}


Point3d    matrix_multiply( double  mCoVars[3][3],
                            Point3d  & vec )
{
    double  sum;
    Point3d  out;

    for  ( int  ind = 0; ind < 3; ind++ ) {
        sum = 0;
        for  ( int  jnd = 0; jnd < 3; jnd++ )
            sum += mCoVars[ ind ][ jnd ] * vec[ jnd ];

        out[ ind ] = sum;
    }

    return   out;
}


void MakeCovarOBB( const ArrPoint3d  & arr,
                   Point3d  & base_pnt,
                   Point3d  & side_0,
                   Point3d  & side_1,
                   Point3d  & side_2 )
{
    //double  boxpt[3], boxvecs[3][3];
    double  oon;
    double  mCoVars[3][3], mXForm[3][3];
    //double  ptAvg[3], ptAvgSq[3], ptAvgXS[3];
    Point3d  ptAvg, ptAvgSq, ptAvgXS;
    double  EigenVals[3];
    double  XfMin[3], XfMax[3];
    double  XfTemp;
    //int  i;
    
    assert( arr.size() > 0 );

    oon = 1.0 / (double)arr.size();

    /* Make averages for covariance matrix. */

    /* Intitialize averages. */
    ptAvg.zero();
    ptAvgSq.zero();
    ptAvgXS.zero();
   
    /* Sum all point data. */
    for  ( int  i = 0; i < arr.size(); i++ ) {
        ptAvg.add( arr[ i ] );

        ptAvgSq[X] += arr[i][X] * arr[i][X];
        ptAvgSq[Y] += arr[i][Y] * arr[i][Y];
        ptAvgSq[Z] += arr[i][Z] * arr[i][Z];

        ptAvgXS[X] += arr[i][X]*arr[i][Y];
        ptAvgXS[Y] += arr[i][X]*arr[i][Z];
        ptAvgXS[Z] += arr[i][Y]*arr[i][Z];
    }

    /* Divide sums to make average. */
    ptAvg.scale( oon );
    ptAvgSq.scale( oon );
    ptAvgXS.scale( oon );

    /* Compose covariance matrix. */
    mCoVars[0][0] = ptAvgSq[X] - ptAvg[X]*ptAvg[X];
    mCoVars[0][1] = ptAvgXS[X] - ptAvg[X]*ptAvg[Y];
    mCoVars[0][2] = ptAvgXS[Y] - ptAvg[X]*ptAvg[Z];

    mCoVars[1][0] = ptAvgXS[X] - ptAvg[Y]*ptAvg[X];
    mCoVars[1][1] = ptAvgSq[Y] - ptAvg[Y]*ptAvg[Y];
    mCoVars[1][2] = ptAvgXS[Z] - ptAvg[Y]*ptAvg[Z];

    mCoVars[2][0] = ptAvgXS[Y] - ptAvg[Z]*ptAvg[X];
    mCoVars[2][1] = ptAvgXS[Z] - ptAvg[Z]*ptAvg[Y];
    mCoVars[2][2] = ptAvgSq[Z] - ptAvg[Z]*ptAvg[Z];


    /*
    for  ( int  ind = 0; ind < 3; ind++ ) {
        printf( "[ " );
        for  ( int  jnd = 0; jnd < 3; jnd++ )
            printf( "%g ", mCoVars[ ind ][ jnd ] );
        printf( "]\n" );
    }
    */

    /* Find eigenvectors for the covariance matrix. */

    /* put eigenvectors into mXForm */
    //EigenSymmetric3(mCoVars, EigenVals, mXForm);
    Jacobi_EigenSymmetric3( mCoVars, EigenVals, mXForm );
    /*
      printf( "PCA eigen values: %g %g %g\n",
            EigenVals[ 0 ],
            EigenVals[ 1 ],
            EigenVals[ 2 ] );
    */


    /* Normalize the eigenvectors to make unit axes. */
    Point3d  evec[ 3 ];
    for  ( int  i = 0; i < 3; i++ ) {
        //XfTemp = (double)(1.0/sqrt(mXForm[i][0]*mXForm[i][0] +
        //                          mXForm[i][1]*mXForm[i][1] +
        //                          mXForm[i][2]*mXForm[i][2]));
        //mXForm[i][0] *= XfTemp;
        //mXForm[i][1] *= XfTemp;
        //mXForm[i][2] *= XfTemp;
        evec[ i ] = Point3d( mXForm[i][0], 
                             mXForm[i][1],
                             mXForm[i][2] );
        //printf( "vec:%d: ", i );
        //evec[ i ].print();
        //printf( "\n" );
        /*
        Point3d  evec_out;

        evec_out = multiply( mCoVars, evec[ i ] );
        printf( "mult:   " );
        evec_out.print();
        printf( "ratio: %g\n", evec_out.length() / evec[ i ].length() );
        */
    }


    /* find minimum/maximum of points in new coord
       system defined by the eigenvectors */

    /* initialize min & max with first point's values */
    XfMax[ 0 ] = XfMin[ 0 ] = evec[ 0 ].dot_prod( arr[ 0 ] );
    XfMax[ 1 ] = XfMin[ 1 ] = evec[ 1 ].dot_prod( arr[ 0 ] );
    XfMax[ 2 ] = XfMin[ 2 ] = evec[ 2 ].dot_prod( arr[ 0 ] );

    /* check subsequent points */
    for  ( int ind = 1; ind < arr.size(); ind++){
        XfTemp = evec[ 0 ].dot_prod( arr[ ind ] );
        XfMin[ 0 ] = min( XfMin[ 0 ], XfTemp );
        XfMax[ 0 ] = max( XfMax[ 0 ], XfTemp );

        XfTemp = evec[ 1 ].dot_prod( arr[ ind ] );
        XfMin[ 1 ] = min( XfMin[ 1 ], XfTemp );
        XfMax[ 1 ] = max( XfMax[ 1 ], XfTemp );

        XfTemp = evec[ 2 ].dot_prod( arr[ ind ] );
        XfMin[ 2 ] = min( XfMin[ 2 ], XfTemp );
        XfMax[ 2 ] = max( XfMax[ 2 ], XfTemp );
    }

    /* Now the orthogonal bounding box in transformed
       coordinates is in XfMin and XfMax. */
    base_pnt = compute_pnt( evec, XfMin );
    
    side_0 = evec[ 0 ];
    side_0.scale( XfMax[ 0 ] - XfMin[ 0 ] );

    side_1 = evec[ 1 ];
    side_1.scale( XfMax[ 1 ] - XfMin[ 1 ] );

    side_2 = evec[ 2 ];
    side_2.scale( XfMax[ 2 ] - XfMin[ 2 ] );    
}


void  computeParallelogram(  const ArrPoint3d  & arr,
                             const Point3d  & vec1,
                             const Point3d  & vec2,
                             const Point3d  & vec3,
                             Point3d  & base_pnt,
                             Point3d  & side_0,
                             Point3d  & side_1,
                             Point3d  & side_2 )
{
    double  XfMin[3], XfMax[3], XfTemp;

    /* initialize min & max with first point's values */
    XfMax[ 0 ] = XfMin[ 0 ] = vec1.dot_prod( arr[ 0 ] );
    XfMax[ 1 ] = XfMin[ 1 ] = vec2.dot_prod( arr[ 0 ] );
    XfMax[ 2 ] = XfMin[ 2 ] = vec3.dot_prod( arr[ 0 ] );

    /* check subsequent points */
    for  ( int ind = 1; ind < arr.size(); ind++){
        XfTemp = vec1.dot_prod( arr[ ind ] );
        XfMin[ 0 ] = min( XfMin[ 0 ], XfTemp );
        XfMax[ 0 ] = max( XfMax[ 0 ], XfTemp );

        XfTemp = vec2.dot_prod( arr[ ind ] );
        XfMin[ 1 ] = min( XfMin[ 1 ], XfTemp );
        XfMax[ 1 ] = max( XfMax[ 1 ], XfTemp );

        XfTemp = vec3.dot_prod( arr[ ind ] );
        XfMin[ 2 ] = min( XfMin[ 2 ], XfTemp );
        XfMax[ 2 ] = max( XfMax[ 2 ], XfTemp );
    }

    /* Now the orthogonal bounding box in transformed
       coordinates is in XfMin and XfMax. */
    Point3d  evec[ 3 ];

    evec[ 0 ] = vec1;
    evec[ 1 ] = vec2;
    evec[ 2 ] = vec3;

    base_pnt = compute_pnt( evec, XfMin );
    
    side_0 = vec1;
    side_0.scale( XfMax[ 0 ] - XfMin[ 0 ] );

    side_1 = vec2;
    side_1.scale( XfMax[ 1 ] - XfMin[ 1 ] );

    side_2 = vec3;
    side_2.scale( XfMax[ 2 ] - XfMin[ 2 ] );    
}

/* beginning of MAGIC code */

typedef double REAL;
const REAL epsilon = 1e-06f;

// roots r0 < r1 < r2
#define distinctRoots 0
// root r0
#define singleRoot    1
// roots r0 = r1 < r2
#define doubleRoot01  2
// roots r0 < r1 = r2
#define doubleRoot12  4
// roots r0 = r1 = r2
#define tripleRoot    6

//===========================================================================
int CubicRoots (REAL c0, REAL c1, REAL c2, REAL *r0, REAL *r1, REAL *r2)
{
    // polynomial is L^3-c2*L^2+c1*L-c0

    int maxiter = 50;
    REAL discr, temp, pval, pdval, b0, b1;
    int i;

    // find local extrema (if any) of p'(L) = 3*L^2-2*c2*L+c1
    discr = c2*c2-3*c1;
    if ( discr >= 0.0 ) {
        discr = (REAL)sqrt(discr);
        temp = (c2+discr)/3;
        pval = temp*(temp*(temp-c2)+c1)-c0;
        if ( pval >= 0.0 ) {
            // real root occurs before the positive local maximum
            (*r0) = (c2-discr)/3 - 1;  // initial guess for Newton's methods
            pval = 2*epsilon;
            for (i = 0; i < maxiter && fabs(pval) > epsilon; i++) {
                pval = (*r0)*((*r0)*((*r0)-c2)+c1)-c0;
                pdval = (*r0)*(3*(*r0)-2*c2)+c1;
                (*r0) -= pval/pdval;
            }

            // Other two roots are solutions to quadratic equation
            // L^2 + ((*r0)-c2)*L + [(*r0)*((*r0)-c2)+c1] = 0.
            b1 = (*r0)-c2;
            b0 = (*r0)*((*r0)-c2)+c1;
            discr = b1*b1-4*b0;
            if ( discr < -epsilon )
            {
                // single root r0
                return singleRoot;
            }
            else
            {
                int result = distinctRoots;

                // roots r0 <= r1 <= r2
                discr = (REAL)sqrt(fabs(discr));
                (*r1) = (REAL)0.5*(-b1-discr);
                (*r2) = (REAL)0.5*(-b1+discr);

                if ( fabs((*r0)-(*r1)) <= epsilon )
                {
                    (*r0) = (*r1);
                    result |= doubleRoot01;
                }
                if ( fabs((*r1)-(*r2)) <= epsilon )
                {
                    (*r1) = (*r2);
                    result |= doubleRoot12;
                }
                return result;
            }
        }
        else {
            // real root occurs after the negative local minimum
            (*r2) = temp + 1;  // initial guess for Newton's method
            pval = 2*epsilon;
            for (i = 0; i < maxiter && fabs(pval) > epsilon; i++) {
                pval = (*r2)*((*r2)*((*r2)-c2)+c1)-c0;
                pdval = (*r2)*(3*(*r2)-2*c2)+c1;
                (*r2) -= pval/pdval;
            }

            // Other two roots are solutions to quadratic equation
            // L^2 + (r2-c2)*L + [r2*(r2-c2)+c1] = 0.
            b1 = (*r2)-c2;
            b0 = (*r2)*((*r2)-c2)+c1;
            discr = b1*b1-4*b0;
            if ( discr < -epsilon )
            {
                // single root
                (*r0) = (*r2);
                return singleRoot;
            }
            else
            {
                int result = distinctRoots;

                // roots r0 <= r1 <= r2
                discr = (REAL)sqrt(fabs(discr));
                (*r0) = (REAL)0.5*(-b1-discr);
                (*r1) = (REAL)0.5*(-b1+discr);

                if ( fabs((*r0)-(*r1)) <= epsilon )
                {
                    (*r0) = (*r1);
                    result |= doubleRoot01;
                }
                if ( fabs((*r1)-(*r2)) <= epsilon )
                {
                    (*r1) = (*r2);
                    result |= doubleRoot12;
                }
                return result;
            }
        }
    }
    else {
        // p(L) has one real root
        (*r0) = c0;
        pval = 2*epsilon;
        for (i = 0; i < maxiter && fabs(pval) > epsilon; i++) {
            pval = (*r0)*((*r0)*((*r0)-c2)+c1)-c0;
            pdval = (*r0)*(3*(*r0)-2*c2)+c1;
            (*r0) -= pval/pdval;
        }
        return singleRoot;
    }
}
//---------------------------------------------------------------------------
void Tridiagonal3 (REAL mat[3][3], REAL diag[3], REAL subd[2])
{
    REAL a, b, c, d, e, f, ell, q;

    a = mat[0][0];
    b = mat[0][1];
    c = mat[0][2];
    d = mat[1][1];
    e = mat[1][2];
    f = mat[2][2];

    diag[0] = a;
    subd[2] = (REAL)0;
    if ( fabs(c) >= epsilon ) {
        ell = (REAL)sqrt(b*b+c*c);
        b /= ell;
        c /= ell;
        q = 2*b*e+c*(f-d);
        diag[1] = d+c*q;
        diag[2] = f-c*q;
        subd[0] = ell;
        subd[1] = e-b*q;
        mat[0][0] = (REAL)1; mat[0][1] = (REAL)0; mat[0][2] = (REAL)0;
        mat[1][0] = (REAL)0; mat[1][1] = b; mat[1][2] = c;
        mat[2][0] = (REAL)0; mat[2][1] = c; mat[2][2] = -b;
    }
    else {
        diag[1] = d;
        diag[2] = f;
        subd[0] = b;
        subd[1] = e;
        mat[0][0] = (REAL)1; mat[0][1] = (REAL)0; mat[0][2] = (REAL)0;
        mat[1][0] = (REAL)0; mat[1][1] = (REAL)1; mat[1][2] = (REAL)0;
        mat[2][0] = (REAL)0; mat[2][1] = (REAL)0; mat[2][2] = (REAL)1;
    }
}
//---------------------------------------------------------------------------
void PreEigenvector (REAL diag[3], REAL subd[2], REAL eval, REAL evec[3][3],
                     int *numvec)
{
    if ( fabs(subd[0]) >= epsilon )
    {
        if ( fabs(subd[1]) >= epsilon )
        {
            REAL diff = eval-diag[0];
            evec[(*numvec)][0] = subd[0]*subd[1];
            evec[(*numvec)][1] = subd[1]*diff;
            evec[(*numvec)][2] = diff*(eval-diag[1])-subd[0]*subd[0];
            (*numvec)++;
        }
        else
        {
            evec[(*numvec)][0] = subd[0];
            evec[(*numvec)][1] = eval-diag[0];
            evec[(*numvec)][2] = (REAL)0;
            (*numvec)++;
            if ( fabs(eval-diag[2]) < epsilon )
            {
                evec[(*numvec)][0] = (REAL)0;
                evec[(*numvec)][1] = (REAL)0;
                evec[(*numvec)][2] = (REAL)1;
                (*numvec)++;
            }
        }
    }
    else
    {
        if ( fabs(subd[1]) >= epsilon )
        {
            evec[(*numvec)][0] = (REAL)0;
            evec[(*numvec)][1] = subd[1];
            evec[(*numvec)][2] = eval-diag[1];
            (*numvec)++;
            if ( fabs(eval-diag[0]) < epsilon )
            {
                evec[(*numvec)][0] = (REAL)1;
                evec[(*numvec)][1] = (REAL)0;
                evec[(*numvec)][2] = (REAL)0;
            }
        }
        else
        {
            if ( fabs(eval-diag[0]) < epsilon )
            {
                evec[(*numvec)][0] = (REAL)1;
                evec[(*numvec)][1] = (REAL)0;
                evec[(*numvec)][2] = (REAL)0;
                (*numvec)++;
            }
            if ( fabs(eval-diag[1]) < epsilon )
            {
                evec[(*numvec)][0] = (REAL)0;
                evec[(*numvec)][1] = (REAL)1;
                evec[(*numvec)][2] = (REAL)0;
                (*numvec)++;
            }
            if ( fabs(eval-diag[2]) < epsilon )
            {
                evec[(*numvec)][0] = (REAL)0;
                evec[(*numvec)][1] = (REAL)0;
                evec[(*numvec)][2] = (REAL)1;
                (*numvec)++;
            }
        }
    }
}
//---------------------------------------------------------------------------
void EigenSymmetric3 (REAL mat[3][3], REAL eval[3], REAL evec[3][3])
{
    REAL diag[3], subd[2];
    REAL c0, c1, c2, sum;
    REAL prevec[3][3];  // matrix P, vectors stored as rows, not columns
    int k, row, col, flag, numvec;

    //   Householder reduction T = Q^t M Q
    //     Input:   
    //       mat, symmetric 3x3 matrix M
    //     Output:  
    //       mat, orthogonal matrix Q
    //       diag, diagonal entries of T
    //       subd, subdiagonal entries of T (T is symmetric)

    Tridiagonal3(mat,diag,subd);

    // eigenvalue construction
    c0 = diag[0]*diag[1]*diag[2]-diag[0]*subd[1]*subd[1]
             -diag[2]*subd[0]*subd[0];
    c1 = diag[0]*diag[1]+diag[0]*diag[2]+diag[1]*diag[2]
             -subd[0]*subd[0]-subd[1]*subd[1];
    c2 = diag[0]+diag[1]+diag[2];
    flag = CubicRoots(c0,c1,c2,eval + 0, eval + 1, eval + 2);

    // eigenvector construction
    numvec = 0;
    switch ( flag )
    {
    case distinctRoots:
        PreEigenvector(diag,subd,eval[0],prevec,&numvec);
        PreEigenvector(diag,subd,eval[1],prevec,&numvec);
        PreEigenvector(diag,subd,eval[2],prevec,&numvec);
        break;
    case doubleRoot01:
    case doubleRoot12:
        PreEigenvector(diag,subd,eval[0],prevec,&numvec);
        PreEigenvector(diag,subd,eval[2],prevec,&numvec);
        break;
    case tripleRoot:
        PreEigenvector(diag,subd,eval[0],prevec,&numvec);
        break;
    }

    // transform back to original coordinates by E^t = PQ where the
    // rows of E are the eigenvectors of M
    for (row = 0; row < 3; row++)
    {
        for (col = 0; col < 3; col++)
        {
            sum = (REAL)0;
            for (k = 0; k < 3; k++)
                sum += prevec[row][k]*mat[k][col];
            evec[row][col] = sum;
        }
    }
}
/* end of MAGIC code */

