/*--------------------------------------------------------------------
 * cutting - compute palanr cuttings
 * Copyright (C) 1998 Sariel Har-Peled
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA  02111-1307, USA.
\*--------------------------------------------------------------------*/

/*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
 * sarray.h -
 *    Small/static templated array.
\*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*/

#ifndef  __SARRAY__H
#define  __SARRAY__H

#define  DASSERT( X, Y)   {if  ( !(X) ) { fprintf( stderr, Y ); assert(X); }}

/*----------------------------------------------------------------------
 * a function we need to implement a search.
\*----------------------------------------------------------------------*/

int        searchArrayEntry( const void  * ptr, int  size, int num,
                                    const void  * var );


int        searchTableEntry( const void  * ptr,
                                    int           localOffset,
                                    int           localSize,
                                    int           size,
                                    int           numEnt,
                                    const void  * var );
int        calcEntriesNum( void    * ptr, int      size );


template<class T>
void  * operator  new( size_t x, T  * ptr ) 
{
    //printf( "my new called!: %d\n", x ); 
    (void)x;
    return  (void *)ptr;
}

#define  SARRAY_NO_RANGE_CHECK

typedef  void    ( *ptrDestructFunc)( void  * ptr );
typedef  void    ( *ptrCopyConstructFunc)( void  * ptr, void  * src );

class  GenericArray
{
protected:
    bool        fInitialized;
    int         sizeEntry;
    int         maxSize;
    int         currSize;
    void      * ptrArr;
    int  id;
    ptrDestructFunc  ptrDestructor;
    ptrCopyConstructFunc  ptrCopyConstructor;

    void   dump() 
    {
        fflush( stdout );
        fflush( stderr );
        printf( "sarray dump\n" );
        printf( "   fInitialized: %d\n", fInitialized );
        printf( "   maxSize: %d\n", maxSize );
        printf( "   currSize: %d\n", currSize );
        printf( "   pArr: %p\n", (void *)ptrArr );
        printf( "   id: %d\n", id );
    }

    int   hash_gen( int  size_entry ) const
    {
        int  ind, val, jnd;
        const char  * ptr;

        val = 0;
        for  ( ind = 0; ind < getEntriesNum(); ind++ ) {
            ptr = (((char *)ptrArr) + ind * size_entry);
            for  ( jnd = size_entry-1; jnd >= 0; jnd-- ) {
                val = ((val ^ *ptr) << 3) + *ptr + val;
                ptr++;
            }
        }

        return  val;
    }

    void   term_Range( int  size ) 
    {
        (void)size;
        
        int  ind;
        char  * ptr;

        //printf( "term_Range - terminating range( %d )!\n", size );
        ptr = (char *)ptrArr;
        for  ( ind = 0; ind < size; ind++ ) {
            //intf( "ptrDestructor: %p\n", (void *)ptrDestructor );
            (*ptrDestructor)( ptr );
            ptr += sizeEntry;
        }       
    }


public:
    GenericArray( ptrDestructFunc  _ptrDestructor,
                  ptrCopyConstructFunc  _ptrConstructor )
    {
        ptrDestructor = _ptrDestructor;
        ptrCopyConstructor = _ptrConstructor;
    }

    void  init_funcs( ptrDestructFunc  _ptrDestructor,
                      ptrCopyConstructFunc  _ptrConstructor )
    {
        ptrDestructor = _ptrDestructor;
        ptrCopyConstructor = _ptrConstructor;
    }

    void  truncate( int      pos )
    {
        assert( pos >= 0 );
        if  ( currSize >= pos  &&   pos >= 0 )
            currSize = pos;
    }

    bool  isFull( void ) const
    {
        return   currSize >= maxSize;
    }

    bool  isEmpty( void ) const
    {
        return   currSize <= 0;
    }

    void  initToEmpty( void )
    {
        fInitialized = false;
    }

    int        getEntriesNum( void ) const
    {
        return    currSize;
    }
    int        size( void ) const
    {
        return    currSize;
    }

    void       reset( void )
    {
        term_Range( currSize );
        currSize = 0;
    }
};


template<class T>
class  Array : public GenericArray {
private:
    T         * pArr;
 
public:
    typedef  int        ( * ptrCompFunc )( const T     & a, const T    & b );
    typedef  int        ( * ptrCompFuncExt )( void  * data, const T     & a,
                                              const T    & b );

    static  void    destruct( void  * ptr ) 
    {
        //intf( "destructing: %p\n", (void *)ptr );
        ((T *)ptr)->~T();
    }
    static  void    copyConstruct( void  * ptr, void  * src ) 
    {
        ::new( (T *)ptr )  T( *(T *)src );
    }

                                                               
private:
    int               partitionExt( ptrCompFuncExt  pCompFunc,
                                    void  * data, long  left, 
                                    long  right );
   int               partition( ptrCompFunc  p_comp,
                                long  left, long  right );
   void             internalQSortExt( ptrCompFuncExt  pCompFunc,
                                      void  * data, int  left,
                                      int  right );
   void             internalQSort( ptrCompFunc   pCompFunc,
                                   int  left, int  right );

    /*
   void   termRange( int  size ) 
    {
        int  ind;

        for  ( ind = 0; ind < size; ind++ )
            (pArr + ind)->~T();
        //delete( pArr )  (pArr + ind);
    }*/


public:
    Array( void ) : GenericArray( destruct, copyConstruct )
    {
        sizeEntry = ((char *)(pArr + 1) - (char *)pArr);

        fInitialized = false;
    }

    void             quickSort( ptrCompFunc  pFunc );
    int              binarySearch( const T    & elem, ptrCompFunc  pCompFunc );
    int              binarySearchExt( const T    & elem, ptrCompFunc  pCompFunc,
                                      int  left, int  right );
    void             quickSortExt( ptrCompFuncExt  pCompFuncExt,
                                             void  * data );
    void             quickSortReverse( ptrCompFunc  pCompFunc );
    void             substractSet( Array     & set, ptrCompFunc  pCompFunc );
    void             substractSortedSet( Array      & set, ptrCompFunc  pCompFunc );

#ifdef  NO_THANKS
    void  permut()
    {
        int  newPos, ind;

        for   ( ind = 0; ind < currSize; ind++ ) {
            real  k;

            realRand(k);
            newPos = (to_int(k * (3 * currSize))) % currSize;
            assert( 0 <= newPos  &&  newPos < currSize );
            exchange( pArr[ ind ], pArr[ newPos ] );            
        }
    }
#endif  // NO_THANKS

    int   hash( void ) const
    {
        return  hash_gen( sizeof( T ) );
    }

        
    void  reverse( void )
    {
        int                ind;
        T                  tmp;

/*        printf( "reversing\n" );*/
        for  ( ind = (currSize / 2) - 1; ind >= 0; ind-- ) { 
            tmp = (*this)[ ind ];
            (*this)[ ind ] = (*this)[ currSize - ind - 1 ];
            (*this)[ currSize - ind - 1 ] = tmp;
        }
    }


    void  init( int   entriesNum, int  _id )
    {
        init_funcs( destruct, copyConstruct );
        sizeEntry = ( ((char *)(pArr + 1)) - (char *)pArr);

        id = _id;
        maxSize = entriesNum;
        currSize = 0;

        //pArr = new T[ max( sizeof( T ) * entriesNum, 256U ) ];
        //printf( "Malloc size: %d\n", (int)max( sizeof( T ) * entriesNum, 256U ) );

        pArr = (T *)malloc( max( sizeof( T ) * (2+ entriesNum), 256U ) );
        //printf( "%p malloc: %d\n", pArr, id );
        if  ( pArr == NULL ) {
          /*  DEBUG_SystemPanic( __POS__, "Fail to allocate array" ); */
            return;
        }
        ptrArr = pArr;
        fInitialized = true;
    }

    bool    resize( int        entriesNum )
    {
        int         size, ind;
        T         * pNewArr;
        
        ///printf( "resize - [%d]!\n", id );
        DASSERT( ( ((char *)&(pArr[ 1 ])) - ((char *)&(pArr[ 0 ])) ) 
                 == sizeEntry, "Size failure - not initialized?" );

	size = max( sizeof( T ) * (2 + entriesNum), 256U );
        pNewArr = (T *)malloc( size );

        if  ( pNewArr == NULL ) 
            return  false;
	memset( pNewArr, 0, size );

        //printf( "\tcurrSize: %d\n", currSize );
        for  ( ind = 0; ind < currSize; ind++ )
            ::new( pNewArr + ind )  T( pArr[ ind ] );

        //printf( "do termRange\n" );
        term_Range( currSize );
        free( (void *) pArr );

	pArr = pNewArr;
        ptrArr = pArr;
	maxSize = entriesNum;	

	return  true;
    }
    
    bool       grow( int          size ) 
    {
        if  ( size <= maxSize )
            return  true;
        
        return  resize( 256 + size * 4 );
    }

    const T  & top() const {
        assert( currSize > 0 );
        return  pArr[ currSize - 1 ];
    }


    const T &operator[]( int i )  const
    {
#ifndef SARRAY_NO_RANGE_CHECK
        DASSERT( 0 <= i  &&  i < currSize  &&  fInitialized, "range failure" );
#endif
        return( pArr[ i ] );
    }

    const T &getCellDirect( int i )  const
    {
        return( pArr[ i ] );
    }

    ~Array( void )
    {
        term();
    }

    T &operator[]( int i )
    {
#ifndef  SARRAY_NO_RANGE_CHECK
        DASSERT( 0 <= i  &&  i < currSize  &&  fInitialized, "range failure" );
#endif 
        return( pArr[ i ] );
    }    
    
    void       deleteEntry( const T   & var )
    {
        int                 ind;
        
        ind = searchEntry( var );
        if  ( ind >= 0 ) {
            pArr[ ind ] = pArr[ currSize - 1 ];
            currSize--;
        }
    }


    void       deleteEntryByPos( int       ind )
    {
        DASSERT( 0 <= ind  &&  ind < currSize &&  fInitialized, "ranmge failure" );
        if  ( ind != (currSize - 1) ) {
            pArr[ ind ] = pArr[ currSize - 1 ];
        }
        currSize--;
    }
    
    
    void       moveToFront( int  pos )
    {
        T  tmp;
        int                  ind;
        
        DASSERT( 0 <= pos  &&  pos < currSize &&  fInitialized, 
                 "moveToFront: ranmge failure" );

        tmp = pArr[ pos ];
        for  (  ind = pos; ind > 0; ind-- ) 
            pArr[ ind ] = pArr[ ind - 1 ];
        
        pArr[ 0 ] = tmp;
    }
    
    
    void       deleteEntryByPosAndKeepOrder( int       pos )
    {
        int                  ind;
        
        DASSERT( 0 <= pos  &&  pos < currSize &&  fInitialized, 
                 "ranmge failure" );
        
        for  (  ind = pos; ind < currSize - 1; ind++ ) 
            pArr[ ind ] = pArr[ ind + 1 ];
        
        currSize--;
    }

    
    void       term( void )
    {
        if  ( ! fInitialized )
            return;
        if  ( pArr != NULL ) {
            //delete[]  pArr;
            term_Range( currSize );
            free( (void *)pArr );
            //printf( "%p free\n", pArr );
            pArr = NULL;
            ptrArr = NULL;
        }
        
        maxSize = currSize = 0;
        fInitialized = false;
    }
    
    
    void       simulateTerm( void )
    {
        fInitialized = false;
        pArr = NULL;
        ptrArr = NULL;
    }
    
    
    void       insertQuick( const T   & var )
    {
        pArr[ currSize++ ] = var;
    }


    void       insertEntry( const T   & var )
    {
        //printf( "inserting entry!\n" );
        DASSERT( 0 <= currSize  &&  currSize < maxSize, "insert failurex" );
        //memset( (void *)(pArr + currSize), 0, sizeof( T ) );
        ::new(pArr + currSize ) T( var );
        currSize++;
    }
    

    void       insertEntryUnique( const T   & var )
    {
        int  pos;

        pos = searchEntry( var );
        if  ( pos >= 0 )
            return;
        
        pushx( var );
    }
    

    int        allocEntry()
    {
        if  ( currSize >= maxSize ) {
            if  ( ! resize( 4 * currSize + 1 ) )
                return  -1;
        }
        //printf( "bogi\n" );

        DASSERT( currSize < maxSize, "insert failure" );
        //memset( (void *)&(pArr[ currSize ]), 0, sizeof( T ) );
        currSize++;

        //printf( "bogi\n" );
        
        return  currSize - 1;
    }


    int        allocEntryFast()
    {
        if  ( currSize >= maxSize ) {
            if  ( ! resize( 10 * currSize + 1 ) )
                return  -1;
        }

        return  currSize++;
    }
    
    
    bool       growAndInsert( const T   & var )
    {
        //printf( "growAndInsert!\n" );
        if  ( currSize >= maxSize ) {
            if  ( ! resize( 4 * currSize + 1 ) )
                return  false;
        }
        insertEntry( var );
        
        return  true;
    }
    
    void       push_in_beginning( const T  & var ) {
        assert( growAndInsert( var ) );
        for  ( int  ind = size()-1; ind > 0; ind-- ) 
            (*this)[ ind ] = (*this)[ ind - 1 ];
        
        (*this)[ 0 ] = var;
    }


    void       pushx( const T   & var ) 
    {
        //printf( "pushx!\n" );
        assert( growAndInsert( var ) );
    }
    
    void       pushy( const T   & var ) 
    {
        if  ( currSize >= maxSize ) {
            assert(resize( 4 * currSize + 1 ) );
        }
        insertQuick( var );
    }
    
    const T    & lastEntry( void )  const
    {
        DASSERT( currSize > 0, "Small Array::no last element to look at" );
        return( pArr[ currSize - 1 ] );
    }

    T    & lastEntry( void )
    {
        DASSERT( currSize > 0, "Small Array::no last element to look at" );
        return( pArr[ currSize - 1 ] );
    }

    void        pop( void )
    {
        DASSERT( currSize > 0, "Small Array::poping an empty stack" );
        if  ( currSize > 0 )
            currSize--;
    }

    int        searchEntry( const T   & var )
    {
        return   searchArrayEntry( pArr, sizeof( T ), currSize, &var );
    }

    int        findLocation( const T  & var ) {
        for  ( int  ind = 0; ind < size(); ind++ ) 
            if  ( pArr[ ind ] == var ) 
                return  ind;
        return  -1;
    }

    
};



template<class T, int  maxSize>
class  SmallArray {
private:
    int         currSize;
    T           pArr[ maxSize ];

public:
    SmallArray( void )
    {
        currSize = 0;
    }

    bool  isFull( void ) const
    {
        return   currSize >= maxSize;
    }

    bool  isEmpty( void ) const
    {
        return   currSize <= 0;
    }

    void  init( void )
    {
        currSize = 0;
    }

    void  init( T        * ptr )
    {
        currSize = calcEntriesNum( (void *)ptr, sizeof( T ) );
        memcpy( pArr, ptr, currSize * sizeof( T ) );
    }

    void  flush( void )
    {
        currSize = 0;
    }

    const T    & lastEntry( void )  const
    {
        DASSERT( currSize > 0, "Small Array::no last element to look at" );
        return( pArr[ currSize - 1 ] );
    }

    T    & lastEntry( void )
    {
        DASSERT( currSize > 0, "Small Array::no last element to look at" );
        return( pArr[ currSize - 1 ] );
    }

    const T &operator[]( int i )  const
    {
#ifndef  SARRAY_NO_RANGE_CHECK
        DASSERT( 0 <= i  &&  i < currSize,
                 "Small Array::Range failure" );
#endif
        return( pArr[ i ] );
    }

    const T &getEntry( int i )  const
    {
        DASSERT( 0 <= i  &&  i < currSize,
                 "Small Array::Range failure" );
        return( pArr[ i ] );
    }

    ~SmallArray( void )
    {
        term();
    }

    T &operator[]( int i )
    {
#ifndef  SARRAY_NO_RANGE_CHECK
        DASSERT( 0 <= i  &&  i < currSize,
                 "Small Array::Range failure" );
#endif
        return( pArr[ i ] );
    }

    int        getEntriesNum( void ) const
    {
        return    currSize;
    }

    int        size( void ) const
    {
        return    currSize;
    }

    void       bringToTop( const T    & var )
    {
        int                 ind;

        ind = searchEntry( var );
        if  ( ind >= 0  &&  ind < currSize - 1 )
            exchange( pArr[ ind ], pArr[ currSize - 1 ] );
    }

    void       deleteEntryKeepOrder( const T      var )
    {
        int                   ind;

        ind = searchEntry( var );
        if  ( ind >= 0 ) {
            memcpy( (void *)&(pArr[ ind ]),
                     (void *)&(pArr[ ind + 1 ] ),
                     (currSize - ind - 1 ) * sizeof( T ) );
            currSize--;
        }
    }

    void       deleteEntry( const T   & var )
    {
        int                 ind;

        ind = searchEntry( var );
        if  ( ind >= 0 ) {
            pArr[ ind ] = pArr[ currSize - 1 ];
            currSize--;
        }
    }


    void       deleteEntryByPos( int       ind )
    {
        DASSERT( 0 <= ind  &&  ind < currSize,
                 "Small Array::delete entry Range failure" );
        pArr[ ind ] = pArr[ currSize - 1 ];
        currSize--;
    }

    void        pop( void )
    {
        DASSERT( currSize > 0, "Small Array::poping an empty stack" );
        if  ( currSize > 0 )
            currSize--;
    }

    void       term( void )
    {
        currSize = 0;
    }


    void       insertEntry( const T   & var )
    {
        //DASSERT( currSize < maxSize, "Small Array::Insert failure" );
        pArr[ currSize ] = var;
        currSize++;
    }

    void       push( const T       & var )
    {
        insertEntry( var );
    }

    void       reset( void )
    {
        currSize = 0;
    }

    int        searchEntry( const T   & var )
    {
        return   searchArrayEntry( &pArr[0], sizeof( T ),
                                   currSize, &var );
    }

    int        findLocation( const T  & var ) {
        for  ( int  ind = 0; ind < size(); ind++ ) 
            if  ( pArr[ ind ] == var ) 
                return  ind;
        return  -1;
    }
};


#endif  /* __SARRAY__H */

/*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
 *
 * sarray.h - End of File
\*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*/
